@@ -87,9 +87,10 @@ SEXP R_uncompress(const SEXP data) {
   SEXP ret_data;
 
   unsigned char pos = 0;
-  unsigned char info, block_mode, max_bits, table_bits, curstr_alloc, laststr_alloc;
+  unsigned char info, block_mode, max_bits, table_bits;
+  unsigned long curstr_alloc, laststr_alloc, cachestr_alloc;
   unsigned char** table;
-  unsigned char* ret, * laststr, * curstr;
+  unsigned char* ret, * laststr, * curstr, * cachestr;
   unsigned long table_size, table_alloc, ret_size, ret_alloc, max_entries, curstr_len, laststr_len;
 
   if( TYPEOF(data) != RAWSXP ) {
@@ -118,6 +119,8 @@ SEXP R_uncompress(const SEXP data) {
   curstr = 0;
   curstr_alloc = 0;
   curstr_len = 0;
+  cachestr = 0;
+  cachestr_alloc = 0;
   while(start < end) {
     unsigned long code = read_bits(&start, end, &pos, table_bits);
     if( start > end || (start == end && pos != 0) )
@@ -125,8 +128,14 @@ SEXP R_uncompress(const SEXP data) {
     if( block_mode && code == 256 ) {
       reinit_table(&table, block_mode, &table_size, &table_alloc);
       table_bits = 9;
-      if( laststr_alloc )
-        free(laststr);
+      if( laststr_alloc ) {
+        if( !cachestr ) {
+          cachestr = laststr;
+          cachestr_alloc = laststr_alloc;
+        } else {
+          free(laststr);
+        }
+      }
       laststr = 0;
       laststr_alloc = 0;
       laststr_len = 0;
@@ -156,25 +165,40 @@ SEXP R_uncompress(const SEXP data) {
           error_return("corrupt compressed data detected in uncompress() [repeat code issued without prior token]");
         }
 
+        ++curstr_len;
         if( curstr_alloc ) {
-          curstr = realloc(curstr, curstr_len + 1);
+          if( curstr_len > curstr_alloc ) {
+            curstr_alloc <<= 1;
+            curstr = realloc(curstr, curstr_alloc);
+          }
           if( laststr == temp )
             laststr = curstr;
         } else {
-          curstr = malloc(curstr_len + 1);
+          if( curstr_len <= cachestr_alloc ) {
+            curstr_alloc = cachestr_alloc;
+            curstr = cachestr;
+            cachestr = 0;
+            cachestr_alloc = 0;
+          } else {
+            curstr_alloc = curstr_len<<1;
+            curstr = malloc(curstr_alloc);
+          }
           memcpy(curstr, temp, curstr_len);
-          curstr_alloc = 1;
         }
-        curstr[curstr_len++] = laststr[0];
+        curstr[curstr_len-1] = laststr[0];
       } else {
         curstr_len = table[code][0] | (((unsigned short)table[code][1])<<8);
         if( curstr_alloc ) {
           if( laststr == curstr ) {
-            laststr = malloc(laststr_len);
-            memcpy(laststr, curstr, laststr_len);
-            laststr_alloc = 1;
+            laststr_alloc = curstr_alloc;
+          } else {
+            if( !cachestr ) {
+              cachestr = curstr;
+              cachestr_alloc = curstr_alloc;
+            } else {
+              free(curstr);
+            }
           }
-          free(curstr);
         }
         curstr = table[code]+2;
         curstr_alloc = 0;
